import{d as R,u as b,b as v,e as c,f as $,g as F,h as i,o as s,i as o,w as e,j as l,_ as B,k as t,l as C,n as D,p as a,q as I,v as V,x as N,y as P,P as _,z as w,A as f,B as m,F as S,C as A,D as E,E as L,G as T}from"./index.4f49b599.js";const j={class:"flow-run__meta"},Z=R({__name:"FlowRun",setup(q){const p=b(),r=v("id"),k=c(()=>{const u=["Logs","Task Runs","Sub Flow Runs"];return T.xl||u.push("Details"),u}),y=$(F.getFlowRun,[r],{interval:5e3}),n=c(()=>y.response);function h(){p.push(f.flowRuns())}return(u,z)=>{const d=i("router-link"),g=i("p-tabs"),x=i("p-layout-well");return s(),o(x,{class:"flow-run"},{header:e(()=>[l(n)?(s(),o(l(B),{key:0,"flow-run":l(n),onDelete:h},null,8,["flow-run"])):t("",!0)]),well:e(()=>[l(n)?(s(),C(S,{key:0},[D("div",j,[a(l(I),{state:l(n).state},null,8,["state"]),a(l(V),{duration:l(n).duration},null,8,["duration"]),a(l(N),{"flow-id":l(n).flowId},null,8,["flow-id"]),l(n).deploymentId?(s(),o(l(P),{key:0,"deployment-id":l(n).deploymentId},null,8,["deployment-id"])):t("",!0)]),a(l(_)),a(d,{to:l(f).radar(l(r)),class:"flow-run__small-radar-link"},{default:e(()=>[a(l(w),{"flow-run-id":l(r),class:"flow-run__small-radar"},null,8,["flow-run-id"])]),_:1},8,["to"]),a(l(_)),a(l(m),{"flow-run":l(n),alternate:""},null,8,["flow-run"])],64)):t("",!0)]),default:e(()=>[a(g,{tabs:l(k)},{details:e(()=>[a(d,{to:l(f).radar(l(r)),class:"flow-run__small-radar-link"},{default:e(()=>[a(l(w),{"flow-run-id":l(r),class:"flow-run__small-radar"},null,8,["flow-run-id"])]),_:1},8,["to"]),a(l(_)),l(n)?(s(),o(l(m),{key:0,"flow-run":l(n)},null,8,["flow-run"])):t("",!0)]),logs:e(()=>[l(n)?(s(),o(l(A),{key:0,"flow-run":l(n)},null,8,["flow-run"])):t("",!0)]),"task-runs":e(()=>[l(n)?(s(),o(l(E),{key:0,"flow-run-id":l(n).id},null,8,["flow-run-id"])):t("",!0)]),"sub-flow-runs":e(()=>[l(n)?(s(),o(l(L),{key:0,"flow-run-id":l(n).id},null,8,["flow-run-id"])):t("",!0)]),_:1},8,["tabs"])]),_:1})}}});export{Z as default};
