import{d as _,f,Y as k,e as a,h as w,o,i as u,w as c,p as y,j as e,Z as $,l as v,$ as x,a0 as Q,F as C,k as b}from"./index.4f49b599.js";const q=_({__name:"WorkQueues",setup(g){const p={interval:3e4},s=f(k.getWorkQueues,[{}],p),r=a(()=>{var n;return(n=s.response)!=null?n:[]}),l=a(()=>s.executed&&r.value.length==0),i=a(()=>s.executed);return(n,t)=>{const d=w("p-layout-default");return o(),u(d,{class:"queues"},{header:c(()=>[y(e($))]),default:c(()=>[e(i)?(o(),v(C,{key:0},[e(l)?(o(),u(e(x),{key:0})):(o(),u(e(Q),{key:1,"work-queues":e(r),onUpdate:t[0]||(t[0]=m=>e(s).refresh()),onDelete:t[1]||(t[1]=m=>e(s).refresh())},null,8,["work-queues"]))],64)):b("",!0)]),_:1})}}});export{q as default};
