import{d as l,u as i,h as d,o as f,i as h,w as a,p as s,j as u,a1 as m,a2 as w,A as t,Y as k,a3 as c}from"./index.4f49b599.js";const y=l({__name:"WorkQueueCreate",setup(Q){const o=i(),n=()=>{o.push(t.workQueues())},_=async r=>{try{const{id:e}=await k.createWorkQueue(r);c("Work queue has been created","success"),o.push(t.workQueue(e))}catch(e){c("Error occurred while creating new work queue","error"),console.error(e)}};return(r,e)=>{const p=d("p-layout-default");return f(),h(p,null,{header:a(()=>[s(u(m))]),default:a(()=>[s(u(w),{onSubmit:_,onCancel:n})]),_:1})}}});export{y as default};
