import{d as _,f,M as y,e as a,h as k,o as n,i as p,w as r,p as v,j as e,R as x,l as $,S as C,Q as b,F as g,k as h}from"./index.4f49b599.js";const S=_({__name:"Deployments",setup(B){const m={interval:3e4},t=f(y.getDeployments,[{}],m),l=a(()=>{var o;return(o=t.response)!=null?o:[]}),c=a(()=>t.executed&&l.value.length===0),d=a(()=>t.executed);return(o,s)=>{const u=k("p-layout-default");return n(),p(u,{class:"deployments"},{header:r(()=>[v(e(x))]),default:r(()=>[e(d)?(n(),$(g,{key:0},[e(c)?(n(),p(e(C),{key:0})):(n(),p(e(b),{key:1,deployments:e(l),onUpdate:s[0]||(s[0]=i=>e(t).refresh()),onDelete:s[1]||(s[1]=i=>e(t).refresh())},null,8,["deployments"]))],64)):h("",!0)]),_:1})}}});export{S as default};
