import{d as f,u as i,b as p,f as d,g as w,e as m,h as R,o as n,i as s,w as t,j as e,_ as h,k as g,p as v,H as y,A as k}from"./index.4f49b599.js";const A=f({__name:"FlowRunRadar",setup(x){const u=i(),o=p("id"),r={interval:5e3},l=d(w.getFlowRun,[o.value],r),a=m(()=>l.response);function _(){u.push(k.flowRuns())}return(C,F)=>{const c=R("p-layout-full");return n(),s(c,{class:"flow-run-radar"},{header:t(()=>[e(a)?(n(),s(e(h),{key:0,"flow-run":e(a),class:"flow-run-radar__header",onDelete:_},null,8,["flow-run"])):g("",!0)]),default:t(()=>[v(e(y),{"flow-run-id":e(o)},null,8,["flow-run-id"])]),_:1})}}});export{A as default};
