import{d as k,e as y,o as i,l as _,p as t,w as n,F as S,a9 as h,aa as m,j as a,ab as A,ac as C,ad as B,ae as $,a4 as f,af as j,ag as M,ah as O,h as r,i as V,ai as E,aj as F,ak as D}from"./index.4f49b599.js";const L={class:"settings-block"},N=k({__name:"SettingsCodeBlock",props:{engineSettings:null},setup(c){const e=c,s=y(()=>Object.entries(e.engineSettings));return(d,u)=>(i(),_("div",L,[t(a(A),{multiline:""},{default:n(()=>[(i(!0),_(S,null,h(a(s),(l,o)=>(i(),_("div",{key:o,class:"settings-block--code-line"},m(l[0])+": "+m(l[1]),1))),128))]),_:1})]))}});class R extends B($){}const v=C(new R);const q=k({__name:"Settings",async setup(c){let e,s;const d=[{text:"Settings"}],u=([e,s]=f(()=>v.getSettings()),e=await e,s(),e),l=([e,s]=f(()=>v.getVersion()),e=await e,s(),e),o=j(M());return O(()=>D(o.value)),(U,p)=>{const w=r("p-key-value"),g=r("p-label"),b=r("p-layout-default");return i(),V(b,{class:"settings"},{header:n(()=>[t(a(E),{crumbs:d},{actions:n(()=>[t(w,{label:"Version",value:a(l),alternate:""},null,8,["value"])]),_:1})]),default:n(()=>[t(g,{label:"Color Mode",class:"settings__color-mode"},{default:n(()=>[t(a(F),{selected:o.value,"onUpdate:selected":p[0]||(p[0]=x=>o.value=x)},null,8,["selected"])]),_:1}),t(g,{label:"Orion Settings"},{default:n(()=>[t(N,{class:"settings__code-block","engine-settings":a(u)},null,8,["engine-settings"])]),_:1})]),_:1})}}});export{q as default};
