import{d as k,b as D,u as x,f as C,M as R,e as m,h as s,o,i as a,w as n,j as e,T as S,k as u,V as d,p as l,G as V,W as $,X as A,A as B}from"./index.4f49b599.js";const U=k({__name:"Deployment",setup(N){const _=D("id"),y=x(),i={interval:3e5},r=C(R.getDeployment,[_.value],i),t=m(()=>r.response),v=m(()=>t.value?A(t.value.schedule):"");function f(){y.push(B.deployments())}return(O,P)=>{const p=s("p-key-value"),b=s("p-content"),h=s("p-tabs"),w=s("p-layout-well");return o(),a(w,{class:"deployment"},{header:n(()=>[e(t)?(o(),a(e(S),{key:0,deployment:e(t),onUpdate:e(r).refresh,onDelete:f},null,8,["deployment","onUpdate"])):u("",!0)]),well:n(()=>[e(t)?(o(),a(e(d),{key:0,deployment:e(t),alternate:""},null,8,["deployment"])):u("",!0)]),default:n(()=>[e(t)?(o(),a(h,{key:0,tabs:["Overview","Parameters"]},{overview:n(()=>[l(b,{secondary:""},{default:n(()=>{var c;return[l(p,{label:"Schedule",value:e(v)},null,8,["value"]),l(p,{label:"Location",value:e(t).flowData.blob},null,8,["value"]),l(p,{label:"Flow Runner",value:(c=e(t).flowRunner)==null?void 0:c.type},null,8,["value"]),e(V).xl?u("",!0):(o(),a(e(d),{key:0,deployment:e(t)},null,8,["deployment"]))]}),_:1})]),parameters:n(()=>[l(e($),{parameters:e(t).parameters},null,8,["parameters"])]),_:1})):u("",!0)]),_:1})}}});export{U as default};
