import{d as _,f as m,I as d,e as a,h as w,o as t,i as n,w as r,p as k,j as e,J as y,l as x,K as h,L as v,F as C,k as F}from"./index.4f49b599.js";const B=_({__name:"Flows",setup($){const p={interval:3e4},s=m(d.getFlows,[{}],p),l=a(()=>{var o;return(o=s.response)!=null?o:[]}),u=a(()=>s.executed&&l.value.length===0),f=a(()=>s.executed);return(o,c)=>{const i=w("p-layout-default");return t(),n(i,{class:"flows"},{header:r(()=>[k(e(y))]),default:r(()=>[e(f)?(t(),x(C,{key:0},[e(u)?(t(),n(e(h),{key:0})):(t(),n(e(v),{key:1,flows:e(l),onDelete:c[0]||(c[0]=b=>e(s).refresh())},null,8,["flows"]))],64)):F("",!0)]),_:1})}}});export{B as default};
