import{d as C,b as F,u as A,e as l,f as c,I as N,M as S,h as m,o as a,i as p,w as t,j as e,N as g,k as u,O as w,p as _,Q as B,G as I,A as O}from"./index.4f49b599.js";const j=C({__name:"Flow",setup(R){const f=F("id"),d=A(),y=l(()=>{const s=["Deployments"];return I.xl||s.unshift("Details"),s}),i={interval:3e5},b=c(N.getFlow,[f.value],i),o=l(()=>b.response),D=l(()=>({flows:{id:{any_:[f.value]}}})),r=c(S.getDeployments,[D],i),v=l(()=>{var s;return(s=r.response)!=null?s:[]});function k(){d.push(O.flows())}return(s,n)=>{const $=m("p-tabs"),h=m("p-layout-well");return a(),p(h,{class:"flow"},{header:t(()=>[e(o)?(a(),p(e(g),{key:0,flow:e(o),onDelete:k},null,8,["flow"])):u("",!0)]),well:t(()=>[e(o)?(a(),p(e(w),{key:0,flow:e(o)},null,8,["flow"])):u("",!0)]),default:t(()=>[_($,{tabs:e(y)},{details:t(()=>[e(o)?(a(),p(e(w),{key:0,flow:e(o)},null,8,["flow"])):u("",!0)]),deployments:t(()=>[_(e(B),{deployments:e(v),onUpdate:n[0]||(n[0]=x=>e(r).refresh()),onDelete:n[1]||(n[1]=x=>e(r).refresh())},null,8,["deployments"])]),_:1},8,["tabs"])]),_:1})}}});export{j as default};
